//@ts-check
"use strict";

/** @typedef Area
@type {{
  name: string;
  rect: PIXI.Rectangle;
  weather: string;
}}
*/

/**
@typedef World
@type {{
  name: string;
  width: number;
  height: number;
  mapAnimationTime: number;

  pixiRoot: PIXI.Container;
  pixiLayers: { [prefix: string]: WorldLayer };
  pixiShadowsLayer: PIXI.Container;

  layers: { name: string, data: number[] }[];
  targets: { [name: string]: { x: number; y: number; } };
  baddieSpawners: BaddieSpawner[];
  baddies: Baddie[];
  boat: Boat;
  items: Item[];
  npcs: NPC[];
  bullets: Bullet[];
  fxs: FX[];
  getEntityChunkFor: (tileX: number, tileY: number) => LogicChunk;
  areas: Area[]
}}

@typedef WorldLayer
@type {{ tileLayersContainer: PIXI.Container; entitiesContainer: PIXI.Container; particlesContainer: PIXI.Container; }}

@typedef LogicChunk
@type {{
  teleporters: { rect: PIXI.Rectangle; worldName: string; targetName: string; }[];
  plants: Plant[]
}}

@typedef Plant
@type {{
  pos: PIXI.Point,
  growthCountdown: number,
  plantDesc: PlantDesc,
  sprite: PIXI.Sprite,
  spritesheet: Spritesheet,
}}
 */

const worldLayerPrefixes = ["water", "ground", "above"];

const horizontalFlipFlag = 1 << 31;
const verticalFlipFlag = 1 << 30;
const diagonalFlipFlag = 1 << 29;

function loadWorld(worldName) {
  const mapJson = getJXONTree(app.loader.resources[`inGame/${mapsFolderName}/${worldName}.tmx`].data).map;

  const logicChunks = {};
  const pixiRoot = new PIXI.Container();

  /** @type {{ [prefix: string]: WorldLayer }} */
  const pixiLayers = {};

  for (const worldLayerPrefix of worldLayerPrefixes) {
    const tileLayersContainer = new PIXI.Container();
    pixiRoot.addChild(tileLayersContainer);
    const entitiesContainer = new PIXI.Container();
    entitiesContainer.sortableChildren = true;
    pixiRoot.addChild(entitiesContainer);
    const particlesContainer = new PIXI.Container();
    pixiRoot.addChild(particlesContainer);

    pixiLayers[worldLayerPrefix] = { tileLayersContainer, entitiesContainer, particlesContainer };
  }

  /** @returns {LogicChunk} */
  function getEntityChunkFor(tileX, tileY) {
    const chunkX = Math.floor(tileX / chunkSize);
    const chunkY = Math.floor(tileY / chunkSize);

    const existingChunk = logicChunks[`${chunkX}_${chunkY}`];
    if (existingChunk != null) return existingChunk;

    // Heavyweight objects are sorted into chunks
    // Ideally baddies should be there too but then we need to track when they move from chunk to chunk
    /** @type {LogicChunk} */
    const newChunk = logicChunks[`${chunkX}_${chunkY}`] = {
      teleporters: [],
      plants: []
    };

    return newChunk;
  }

  /** @type {World} */
  const world = {
    name: worldName,
    width: mapJson["@width"],
    height: mapJson["@height"],
    mapAnimationTime: 0,

    pixiRoot,
    pixiLayers,
    pixiShadowsLayer: null,

    layers: [],
    targets: {},
    baddieSpawners: [],
    baddies: [],
    boat: null,
    items: [],
    npcs: [],
    bullets: [],
    fxs: [],
    getEntityChunkFor,
    areas: []
  };

  const tilesetTexture = app.loader.resources[`inGame/${mapsFolderName}/tileset.png`].texture;

  for (let layerIndex = 0; layerIndex < mapJson.layer.length; layerIndex++) {
    const jsonLayer = mapJson.layer[layerIndex];
    const layer = world.layers[layerIndex] = { name: jsonLayer["@name"], data: [] };

    const [prefix, suffix] = layer.name.split(":");
    if (!worldLayerPrefixes.includes(prefix)) throw new Error(`Unexpected tile layer named ${layer.name}`);

    const pixiLayer = pixiLayers[prefix];
    if (parseInt(suffix, 10) !== pixiLayer.tileLayersContainer.children.length) throw new Error(`Unexpected non-tile layer named ${layer.name}`);

    const layerData = jsonLayer.data.keyValue.split(",");

    const layerVertices = new Float32Array(world.width * world.height * 4 * 2);
    const layerUvs = new Float32Array(world.width * world.height * 4 * 2);
    const layerIndices = new Uint32Array(world.width * world.height * 6);

    for (let y = 0; y < world.height; y++) {
      for (let x = 0; x < world.width; x++) {
        const i = y * world.width + x;

        layerVertices[i * 4 * 2 + 0] = x;
        layerVertices[i * 4 * 2 + 1] = y + 1;
        layerVertices[i * 4 * 2 + 2] = x + 1;
        layerVertices[i * 4 * 2 + 3] = y + 1;
        layerVertices[i * 4 * 2 + 4] = x + 1;
        layerVertices[i * 4 * 2 + 5] = y;
        layerVertices[i * 4 * 2 + 6] = x;
        layerVertices[i * 4 * 2 + 7] = y;

        let tile = world.layers[layerIndex].data[i] = parseInt(layerData[i], 10);
        const tileType = inGame.tileset.tileTypesById[tile - 1];

        if (tileType) {
          const plantDesc = tileTypeToPlantDesc(tileType);
          if (plantDesc) {
            const logicChunk = getEntityChunkFor(x, y);
            logicChunk.plants.push(createPlant(world, { x: x + 0.5, y: y + 0.5 }, plantDesc));
            tile = 0;
          }
        }

        setTileUvs(layerUvs, i, tile);

        layerIndices[i * 6 + 0] = i * 4 + 0;
        layerIndices[i * 6 + 1] = i * 4 + 1;
        layerIndices[i * 6 + 2] = i * 4 + 2;
        layerIndices[i * 6 + 3] = i * 4 + 0;
        layerIndices[i * 6 + 4] = i * 4 + 2;
        layerIndices[i * 6 + 5] = i * 4 + 3;
      }
    }

    const layerMesh = new PIXI.SimpleMesh(tilesetTexture, layerVertices, layerUvs, layerIndices);
    pixiLayer.tileLayersContainer.addChild(layerMesh);

    if (layer.name === "ground:1") world.pixiShadowsLayer = layerMesh;
  }

  // Process object layers
  const objectLayers = Array.isArray(mapJson.objectgroup) ? mapJson.objectgroup : [mapJson.objectgroup];
  for (const objectLayer of objectLayers) {
    const objects = Array.isArray(objectLayer.object) ? objectLayer.object : [objectLayer.object];

    switch (objectLayer["@name"]) {
      case "logic": processLogicLayer(objects); break;
      case "areas": processAreasLayer(objects); break;
    }
  }

  return world;

  function processLogicLayer(layer) {
    for (const object of layer) {
      const pos = { x: Math.floor(object["@x"] / tileSize), y: Math.floor(object["@y"] / tileSize) };
      const entityChunk = getEntityChunkFor(pos.x, pos.y);

      /** @type {string} */
      let objectName = object["@name"];
      if (objectName == "Relative:aunty") {
        objectName = "Npc:grandma";
      }

      if (objectName.startsWith("Target:")) {
        world.targets[objectName.substring("Target:".length)] = pos;

      } else if (objectName.startsWith("Teleporter:")) {
        const [worldName, targetName] = objectName.substring("Teleporter:".length).split(".");
        const rect = new PIXI.Rectangle(object["@x"] / tileSize, object["@y"] / tileSize, object["@width"] / tileSize, object["@height"] / tileSize);
        entityChunk.teleporters.push({ rect, worldName, targetName });

      } else if (objectName.startsWith("Item:")) {
        const itemName = objectName.substring("Item:".length);
        const desc = getItemDesc(itemName);
        if (desc) {
          const item = createItem(world, pos, desc);
          world.items.push(item);
        } else {
          console.warn(`No such item: ${itemName}`);
        }

      } else if (objectName.startsWith("Npc:")) {
        const npcName = objectName.substring("Npc:".length);
        const desc = getNPCDesc(npcName);
        if (desc) {
          createNPC(world, pos, desc);
        } else {
          console.warn(`No such npc: ${npcName}`);
        }

      } else if (objectName.startsWith("Special:boat")) {
        createBoat(world, pos);

      } else if (objectName.startsWith("BaddieSpawner:")) {
        const difficulty = parseInt(objectName.substring("BaddieSpawner:".length), 10);
        world.baddieSpawners.push({ pos, difficulty, countdown: 0, interval: { min: 8, max: 12 } });

      } else throw new Error(`Unexpected object named ${objectName}`);
    }
  }

  function processAreasLayer(objects) {
    for (const object of objects) {
      if (object["@name"] != null && object["@name"].startsWith("Area:")) {
        const rect = new PIXI.Rectangle(object["@x"] / tileSize, object["@y"] / tileSize, object["@width"] / tileSize, object["@height"] / tileSize);
        const [name, weather] = object["@name"].substring("Area:".length).split("|");
        world.areas.push({ rect, name, weather });
      }
    }
  }
}

function setTileUvs(layerUvs, i, tile) {
  if (tile !== 0) {
    const tileUvWidth = inGame.tileset.tileWidth / inGame.tileset.width;
    const tileUvHeight = inGame.tileset.tileHeight / inGame.tileset.height;

    const horizontalFlip = (tile & horizontalFlipFlag) !== 0;
    const verticalFlip = (tile & verticalFlipFlag) !== 0;
    const diagonalFlip = (tile & diagonalFlipFlag) !== 0;

    const tileId = (tile & ~(horizontalFlipFlag | verticalFlipFlag | diagonalFlipFlag)) - 1;

    const tileX = tileId % inGame.tileset.columns;
    const tileU = tileX / inGame.tileset.columns;
    const tileY = Math.floor(tileId / inGame.tileset.columns);
    const tileV = tileY / inGame.tileset.rows;

    const bleedingEps = 0.5 / inGame.tileset.width;
    let left = tileU + bleedingEps;
    let right = tileU + tileUvWidth - bleedingEps;
    let top = tileV + tileUvHeight - bleedingEps;
    let down = tileV + bleedingEps;

    if (diagonalFlip) {
      if (!verticalFlip) [left, right] = [right, left];
      if (horizontalFlip) [top, down] = [down, top];

      layerUvs[i * 4 * 2 + 0] = left;
      layerUvs[i * 4 * 2 + 1] = down;
      layerUvs[i * 4 * 2 + 2] = left;
      layerUvs[i * 4 * 2 + 3] = top;
      layerUvs[i * 4 * 2 + 4] = right;
      layerUvs[i * 4 * 2 + 5] = top;
      layerUvs[i * 4 * 2 + 6] = right;
      layerUvs[i * 4 * 2 + 7] = down;
    } else {
      if (horizontalFlip) [left, right] = [right, left];
      if (verticalFlip) [top, down] = [down, top];

      layerUvs[i * 4 * 2 + 0] = left;
      layerUvs[i * 4 * 2 + 1] = top;
      layerUvs[i * 4 * 2 + 2] = right;
      layerUvs[i * 4 * 2 + 3] = top;
      layerUvs[i * 4 * 2 + 4] = right;
      layerUvs[i * 4 * 2 + 5] = down;
      layerUvs[i * 4 * 2 + 6] = left;
      layerUvs[i * 4 * 2 + 7] = down;
    }
  } else {
    layerUvs[i * 4 * 2 + 0] = 1;
    layerUvs[i * 4 * 2 + 1] = 1;
    layerUvs[i * 4 * 2 + 2] = 1;
    layerUvs[i * 4 * 2 + 3] = 1;
    layerUvs[i * 4 * 2 + 4] = 1;
    layerUvs[i * 4 * 2 + 5] = 1;
    layerUvs[i * 4 * 2 + 6] = 1;
    layerUvs[i * 4 * 2 + 7] = 1;
  }
}

/**
 * @param {string} worldName
 * @param {string} targetName
 */
function setActiveWorld(worldName, targetName) {
  /** @type {NPC[]} */
  let followingNpcs = [];

  isInCombat = false;

  if (inGame.activeWorld != null) {
    // Clear bullets
    const { entitiesContainer } = inGame.activeWorld.pixiLayers["ground"];

    for (const bullet of inGame.activeWorld.bullets) {
      entitiesContainer.removeChild(bullet.sprite);
      bullet.sprite.destroy();
    }
    inGame.activeWorld.bullets.length = 0;

    for (const fx of inGame.activeWorld.fxs) {
      if (fx.animSprite) {
        entitiesContainer.removeChild(fx.animSprite);
        fx.animSprite.destroy();
      } else {
        entitiesContainer.removeChild(fx.sprite);
        fx.sprite.destroy();
      }
    }
    inGame.activeWorld.fxs.length = 0;

    let i = 0;
    while (i < inGame.activeWorld.npcs.length) {
      const npc = inGame.activeWorld.npcs[i];

      if (npc.following) {
        followingNpcs.push(npc);
        inGame.activeWorld.pixiLayers["ground"].entitiesContainer.removeChild(npc.pixiRoot);
        inGame.activeWorld.pixiShadowsLayer.removeChild(npc.shadowSprite);
        inGame.activeWorld.npcs.splice(i, 1);
      } else {
        i++;
      }
    }

    stopBaxtonFootsteps();

    inGame.activeWorld = null;
    inGame.pixiWorldContainer.removeChildren();
  }

  if (worldName != null) {
    inGame.activeWorld = inGame.worlds[worldName];
    inGame.pixiWorldContainer.addChild(inGame.activeWorld.pixiRoot);
    inGame.pixiWorldContainer.addChild(inGame.dialog.pixiRoot);

    const target = inGame.activeWorld.targets[targetName];
    if (target == null) throw new Error(`Target "${targetName}" is missing, can't spawn player`);

    inGame.stats.hovered = null;
    inGame.baxton.pos.set(target.x + 0.5, target.y + 0.5);
    inGame.activeWorld.pixiLayers["ground"].entitiesContainer.addChild(inGame.baxton.pixiRoot);
    inGame.activeWorld.pixiShadowsLayer.addChild(inGame.baxton.shadowSprite);

    for (let npc of followingNpcs) {
      npc.pos.copyFrom(inGame.baxton.pos);

      inGame.activeWorld.pixiLayers["ground"].entitiesContainer.addChild(npc.pixiRoot);
      inGame.activeWorld.pixiShadowsLayer.addChild(npc.shadowSprite);
      inGame.activeWorld.npcs.push(npc);
    }

    inGame.baxton.area = null;
    setupBaxtonArea(false);

    inGame.camera.pos.copyFrom(inGame.baxton.pos);

    startFadeIn();
    showWorldTitle(worldName, targetName);
  }
}

/** @param {number} dt */
function animateWorldTiles(dt) {
  const animTime = inGame.activeWorld.mapAnimationTime += dt;

  const cameraX = Math.floor(inGame.camera.pos.x);
  const cameraY = Math.floor(inGame.camera.pos.y);

  for (let layerIndex = 0; layerIndex < inGame.activeWorld.layers.length; layerIndex++) {
    const layer = inGame.activeWorld.layers[layerIndex];

    const [prefix, suffix] = layer.name.split(":");

    /** @type {PIXI.SimpleMesh} */
    const mesh = (/** @type {PIXI.SimpleMesh} */ inGame.activeWorld.pixiLayers[prefix].tileLayersContainer.getChildAt(parseInt(suffix, 10)));
    const uvsBuffer = mesh.geometry.getBuffer("aTextureCoord");
    let hasChangedUvs = false;

    for (let y = Math.max(cameraY - 6, 0); y < Math.min(cameraY + 6, inGame.activeWorld.height); y++) {
      for (let x = Math.max(cameraX - 12, 0); x < Math.min(cameraX + 12, inGame.activeWorld.width); x++) {
        const i = y * inGame.activeWorld.width + x;
        const tile = layer.data[i];

        const tileId = (tile & ~(horizontalFlipFlag | verticalFlipFlag | diagonalFlipFlag)) - 1;
        const tileAnimation = inGame.tileset.animationsById[tileId];
        if (tileAnimation == null) continue;

        hasChangedUvs = true;

        const horizontalFlip = (tile & horizontalFlipFlag) !== 0;
        const verticalFlip = (tile & verticalFlipFlag) !== 0;
        const diagonalFlip = (tile & diagonalFlipFlag) !== 0;

        const timeInAnimation = animTime % tileAnimation.totalDuration;
        let tmpTime = 0;

        for (const frame of tileAnimation.frames) {
          tmpTime += frame.duration;
          if (timeInAnimation <= tmpTime) {
            let tile = frame.tileId + 1;
            if (horizontalFlip) tile += horizontalFlipFlag;
            if (verticalFlip) tile += verticalFlipFlag;
            if (diagonalFlip) tile += diagonalFlipFlag;

            setTileUvs(uvsBuffer.data, i, tile);
            break;
          }
        }
      }
    }

    // @ts-ignore
    if (hasChangedUvs) uvsBuffer._updateID++;
  }
}

/**
 * @param {string} tileType
 * @returns {PlantDesc}
 */
function tileTypeToPlantDesc(tileType) {
  let prefix = "Plant:";

  if (!tileType.startsWith(prefix)) {
    return undefined;
  }

  const numberPart = tileType.slice(prefix.length);
  switch (numberPart) {
    case "0": return plantDescs.leaf;
    case "1": return plantDescs.flower;
    case "2": return plantDescs.dragonfly;
    case "3": return plantDescs.eyes;
    case "4": return plantDescs.nose;
    default: throw new Error(`Unknown plant number: ${numberPart}`);
  }
}

/**
 * @param {string} name
 * @returns {NPCDesc | undefined}
 */
function getNPCDesc(name) {
  return npcDescs[name];
}

/**
 * @param {string} name
 * @returns {ItemDesc | undefined}
 */
function getItemDesc(name) {
  return itemDescs[name];
}

/**
 * @param {NPCName} name
 * @returns {NPC | null}
 */
function findNPC(name) {
  for (const npc of inGame.activeWorld.npcs) {
    if (npc.desc.id == name) {
      return npc;
    }
  }
  return null;
}
